# Data Structures

In this section, we will learn 3 essential data structures concepts:

* Variables store single values, such as numbers, text, or other values. 
* Lists (or arrays) store ordered collections of values, allowing access by index. They help manage multiple values efficiently.
* Dictionaries (or hashmaps) store key-value pairs, allowing data to be retrieved using meaningful keys instead of numerical indices, making lookups more intuitive.

## Variable

### Basics
{%
    include-markdown "../bash/bash-2-the-basics.md"
    start="<!--variables-start-->"
    end="<!--variables-end-->"
%}

### Using Variables


| command | explanation |
|----------|----------|
| **variable=value** | Assign a value `value` to the variable `variable`. (The variable scipe is restricted to the shell/script) |
| **local variable=value** | Assign a value `value` to the variable `variable`. (It does not come out a curly bracket area) |
| **export varialbe=value** | Make the variable `variable` available to the shell abd syb-proessses |
| **variable=$(command)** | Assign the output of `command` to `variable`|
| **{% raw %}${#variable}{% endraw %}** | Length of the value contained by the variable |
| **${variable:N}** | Kepp the character of the value contained by `variable` after the Nth. |
| **$variable:N:length** | Substring the value contained by `variable` from tje Nth chracter up to the `length` specified |
| **${variable/pattern/string}** | Match of `pattern` in `variable` value is replaced with `string` |


## List (Indexed Array)

Lists (or arrays) store ordered collections of values, allowing access by index. They help manage multiple values efficiently.


### create - initialize

This is not mandatory but it is a good practice to create and/or initialize an array!

```bash
declare -a myarray
```
This will declare an array called `myarray`, but it is not initialized.  

```bash
myarray=()
```
This will declare an array called `myarray` and initialize it as an empty array.

```bash
myarray=(apple banana "cherry cake")
```
This `myarray` array is initialized with three values.

### append

In Bash, you can add a value to an array using the following syntax:

```bash
myarray+=("kiwi")
```

### access

To check the size of your array (print the last index):

{% raw %}```python
echo ${#myarray[@]}
```{% endraw %}

To print all array index key :

{% raw %}```python
echo ${!myarray[@]}
```{% endraw %}

**/!\ Index start at 0.**

To print the value stored in index 1 :  
```bash
echo ${myarray[1]}
```


To print all values stored in an array :

{% raw %}```bash
echo ${myarray[@]}
```{% endraw %}

### remove

To remove element at index 1 from the array:
{% raw %}```bash
unset -v myarray[1]
```{% endraw %}

To remove the whole array:
{% raw %}```bash
unset -v myarray
```{% endraw %}

## Dictionary (Associative array)

Associative arrays were introduced in Bash 4.0, which was released in 2009.
It significantly increases the flexibility of working with arrays in Bash scripts, but keep in mind it is slow.
Avoid big associative array, and use it by parsimony.

### create - initialize

To use an associative array you **must** declare it first!  
Then initialize with value is not mandatory.

```bash
declare -A myAArray
```

```bash
myAArray["name"]="John"
```
This `myAArray` array is initialized with 1 values.

```bash
myAArray=([fruit]=apple [cake]="Paris brest")
```
This `myAArray` array is initialized with 2 values. Note the parenthesis!

### append

In Bash, you can add a value to an associative array using the following syntax :

```bash
myAArray["name"]="John"
```

### access

To check the number of value/key in your associative array :

{% raw %}```bash
echo ${#myAArray[@]}
```{% endraw %}

To print all associative array keys :

{% raw %}```bash
echo ${!myAArray[@]}
```{% endraw %}

To print the value associated to a specific key (e.g. cake) :

{% raw %}```bash
echo "${myAArray[cake]}"
```{% endraw %}

To print all values stored in an associated array :

{% raw %}```bash
echo ${myAArray[@]}
```{% endraw %}

### remove

To remove a specific key-value pair the associative array:

```bash
unset -v myAArray[cake]
```

To remove the whole associative array:

```bash
unset -v myAArray
```

!!! Success "Quick recap"
    In this section we've learned:

    - What is variable
    - What is an array (think as list)
    - What is an associative array (think as dictionnary)