# User Interface 

In Bash, user interaction is important for dynamic and flexible scripts.  
Bash provides built-in commands like:

* `read` - To take user input.
* `select` - To create menus.


## Read

The Bash read command is a built-in utility that reads text from standard input

```bash
#!/bin/bash

# first example
echo "Enter your name:"
read name
echo "Hello, $name!"

# second example
read -p "Enter your age: " age
echo "You are $age years old."
```

## Select

The select command is great for interactive menus.

Here with a case statement:  
```bash
#!/bin/bash

#custom prompt for the select 
echo "Choose a cake:"

select cake in Eclair Macaron Millefeuille Quit
do
    case $cake in
        Macaron) echo "You chose Macaron." ;;
        Eclair) echo "You chose Eclair." ;;
        Millefeuille) echo "You chose Millefeuille." ;;
        Quit) echo "Exiting..."; break ;;
        *) echo "Invalid option, try again." ;;
    esac
done
```

Explanation:  

* The select loop automatically displays a numbered menu.
* The user selects by typing a number (e.g., 1 for Apple).
* Quit option allows exiting the menu.


And with a if statements:  
```bash
#!/bin/bash

#custom prompt for the select 
PS3="Choose a cake:"

select opt in Eclair Macaron Millefeuille Quit;do
    if [[ $opt == "Eclair" ]];then
        echo "You chose Eclair." ;
    elif [[ $opt == "Macaron" ]];then
        echo "You chose Macaron." ;
    elif [[ $opt == "Millefeuille" ]];then
        echo "You chose Millefeuille." ;
    elif [[ $opt == "Quit" ]];then
        echo "Exiting..."; break ;
    fi
done
```